/* sendip.c
	Sends specified ip packets
*/

/* Command line options:
	./sendip <hostname> -p <type> -d <data> <options>
	
	<hostname> is the name of the host to send the packet to, defaults to
	localhost
	<type> is the packet type: currently UDP, TCP or ICMP, defaults to IP
	<data> is the data payload of the packet as a string of hexadecimal
	numbers, defaults to 10 bytes of 0s
	
	OPTION     DESCRIPTION     DEFAULT
	Socket options
	-sd        Debug           0
	-sr        Don't route     0 (ie do route)
	-sb        Allow broadcast 0
	
	IP options:
	-is        Source IP       127.0.0.1
	-id        Destination IP  Correct
	-ih        Header length   Correct
	-iy        Type of service 0 (see RFC791 et al)
	-il        Length          Correct
	-ii        Identification  Random
	-ifr       Reserved flag   0
	-ifd       Don't fragment  0
	-ifm       More fragments  0
	-if        Frag offset     0
	-it        TTL             255
	-ip        Protocol        Correct for type
	-ic        Checksum        Correct
	-io        Options - not yet implemented
	
	ICMP options:
	-ct        Type            8 (See RFC792)
	-cd        Code            0 (See RFC792)
	-cc        Checksum        Correct
	
	UDP options:
	-us        Source port      0
	-ud        Destination port 0
	-ul        Length          Correct
	-uc        Checksum        Correct
	
	TCP options:
	-ts        Source port      0
	-td		  Destination port 0
	-tn        Sequence number Random
	-ta        Acknowledgement 0
	-tt        Data offset     Correct
	-tr        Reserved        0
	-tfu       Set urg bit     0 unless -tu is specified
	-tfa       Set ack bit     0 unless -ta is specified
	-tfp       Set psh bit     0
	-tfr       Set rst bit     0
	-tfs       Set syn bit     1
	-tff       Set fin bit     0
	-tw        Window          65535
	-tc        Checksum        Correct
	-tu        Urgent pointer  0
	-to        Options - not yet implemented
*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <arpa/inet.h>
#include <rpc/types.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include "protocols.h"

/* IP_MAXPACKET is not defined on pre-glibc boxes so define it in case */
#ifndef IP_MAXPACKET
#define IP_MAXPACKET 65535
#endif

void do_sockopt(int s, char *opt, char *val);
void usage_exit(const char *err);

const char *progname;

/* print usage message and exit */
void usage_exit(const char *err) {
	fprintf(stderr,"Usage: %s <hostname> <options>\n",progname);
	fprintf(stderr,err);
	fprintf(stderr,"See sendip.c for more information\n");
	exit(1);
}

/* handle socket options */
void do_sockopt(int s, char *opt, char *val) {
	const int zero = 0;
	const int one = 1;
	int r;

	if(*val != '0' && *val != '1') {
		close(s);
		usage_exit("Value for socket option must be 0 or 1\n");
	}
	*val-=0x30;

	switch(opt[2]) {
	case 'd':
	case 'D':
		if(*val)
			r = setsockopt(s, SOL_SOCKET, SO_DEBUG, &one, sizeof(one));
		else
			r = setsockopt(s, SOL_SOCKET, SO_DEBUG, &zero, sizeof(zero));
		if(r)
			perror("Warning: setsockopt(SO_DEBUG)");
		break;
	case 'r':
	case 'R':
		if(*val)
			r = setsockopt(s, SOL_SOCKET, SO_DONTROUTE, &one, sizeof(one));
		else
			r = setsockopt(s, SOL_SOCKET, SO_DONTROUTE, &zero, sizeof(zero));
		if(r)
			perror("Warning: setsockopt(SO_DONTROUTE)");
		break;
	case 'b':
	case 'B':
		if(*val)
			r = setsockopt(s, SOL_SOCKET, SO_BROADCAST, &one, sizeof(one));
		else
			r = setsockopt(s, SOL_SOCKET, SO_BROADCAST, &zero, sizeof(zero));
		if(r)
			perror("Warning: setsockopt(SO_DONTROUTE)");
		break;
	default:
		close(s);
		usage_exit("Unknown socket option\n");
		break;
	}
}	


/* Main program code */
int main(int argc, char **argv) {
	/* socket stuff */
	int s;                            /* socket for sending   */

	/* hostname stuff */
	char hnamebuf[MAXHOSTNAMELEN];    /* buffer for hostname  */
	char *hostname = NULL;            /* target hostname      */
	struct hostent *hp;               /* hostname             */
	struct sockaddr whereto;          /* destination          */
	struct sockaddr_in *to = (struct sockaddr_in *)&whereto;
	                                  /*             address  */

	/* packet stuff */
	unsigned char outpack[IP_MAXPACKET];
                                     /* actual output packet */
	struct iphdr *iph=(struct iphdr *)outpack;
	                                  /* IP packet header     */
	unsigned char *outp;

	struct icmphdr *icp=NULL;         /* ICMP packet header   */
	struct tcphdr *tcp=NULL;          /* TCP packet header    */
	struct udphdr *udp=NULL;          /* UDP packet header    */
	char *data = NULL;                /* packet data argument */

	int sent;                         /* number of bytes sent */

	/* Variable initializations */
	progname=argv[0];
   memset(to, 0, sizeof(struct sockaddr));
	memset(outpack, 0, IP_MAXPACKET);
	memset(hnamebuf, 0, MAXHOSTNAMELEN);
	outp = outpack+sizeof(struct iphdr);
	ip_initialize(iph);

	/* Process command line options - first make sure we have a host name */
	if(argc < 2 || argv[1][0]=='-') usage_exit("No hostname specified\n");
	 
	/* handle hostname - need to end up with whereto initialized */
   if(inet_aton(argv[1], &to->sin_addr)) { /* Esay case - we are given an IP */
		to->sin_family = AF_INET;
      hostname = argv[1];
   } else {                                /* Got a host name */
      hp = gethostbyname(argv[1]);
      if(hp == NULL) {
			perror("gethostbyname");
			return 2;
      }
      to->sin_family = hp->h_addrtype;
      if (hp->h_length > (int)sizeof(to->sin_addr)) { /* make sure it'll fit */
         hp->h_length = sizeof(to->sin_addr);
      }
      memcpy(&to->sin_addr, hp->h_addr, hp->h_length); /* copy address and */
      (void)strncpy(hnamebuf, hp->h_name, sizeof(hnamebuf) - 1); /* name   */
      hostname = hnamebuf;
   }
	
	/* Now we can open the socket before handling the rest of the command line
		options */
	if((s = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
		perror("Couldn't open RAW socket");
		return 2;
	}

	/* Now process the rest of the command line options */
	argc-=2;
	argv+=2;

	while(argc > 1) {
		if(argv[0][0]!='-') {
			close(s);
			usage_exit("Option not starting with a -\n");
		}
		switch(argv[0][1]) {
		case 'c':  /* icmp options */
			if(icp)
				do_icmpopt(icp,argv[0],argv[1]);
			else {
				close(s);
				usage_exit("Can not specify -c* without -p ICMP\n");
			}
			break;
		case 'd':  /* packet data */
			data=argv[1];
			break;
		case 'i':  /* ip options */
			do_ipopt(iph,argv[0],argv[1]);
			break;
		case 'p':  /* packet type */
			if(icp || udp || tcp) {
				close(s);
				usage_exit("Can't specify multiple protocols\n");
			}
			switch(argv[1][0]) {
			case 'I':
			case 'i':
				if(strcasecmp(argv[1],"icmp")) {
					close(s);
					usage_exit("Protocol unknown\n");
				}
				icp=(struct icmphdr *)outp;
				icmp_initialize(icp);
				outp+=4; /*sizoef(struct icmphdr) includes too much gubbins */;
				break;
			case 'U':
			case 'u':
				if(strcasecmp(argv[1],"UDP")) {
					close(s);
					usage_exit("Protocol unknown\n");
				}
				udp=(struct udphdr *)outp;
				udp_initialize(udp);
				outp+=sizeof(struct udphdr);
				break;
			case 'T':
			case 't':
				if(strcasecmp(argv[1],"TCP")) {
					close(s);
					usage_exit("Protocol unknown\n");
				}
				tcp=(struct tcphdr *)outp;
				tcp_initialize(tcp);
				outp+=sizeof(struct tcphdr);
				break;
			default:
				close(s);
				usage_exit("Protocol unknown\n");
				break;
			}
			break;  /* end of packet type */
		case 's':  /* socket options */
			do_sockopt(s,argv[0],argv[1]);
			break;
		case 't':  /* tcp options */
			if(tcp)
				do_tcpopt(tcp,argv[0],argv[1]);
			else {
				close(s);
				usage_exit("Can not specify -t* without -p TCP\n");
			}
			break;
		case 'u':  /* udp options */
			if(udp)
				do_udpopt(udp,argv[0],argv[1]);
			else {
				close(s);
				usage_exit("Can not specify -u* without -p UDP\n");
			}
			break;
		default:  /* something else */
			close(s);
			usage_exit("Unknown option\n");
			break;
		}
		argc-=2; argv+=2;
	}
	if(argc==1) {
		close(s);
		usage_exit("Argument needs a parameter\n");
	}

	/* finalize packet for sending - first, copy data */
	/* Add data from outp - this code is foul */
	if(data) {
		for(;outp-outpack<IP_MAXPACKET && *data;data++, outp++) {
			if(isdigit(*data)) {
				*outp=(*data)-'0';
			} else if('A' <= *data && *data <= 'F') {
				*outp=(*data)-'A'+0x0a;
			} else if('a' <= *data && *data <= 'f') {
				*outp=(*data)-'a'+0x0a;
			}
			(*outp)*=0x10;
			data++;
			if(*data) {
				if(isdigit(*data)) {
				  *outp+=(*data)-'0';
				} else if('A' <= *data && *data <= 'F') {
					*outp+=(*data)-'A'+0x0a;
				} else if('a' <= *data && *data <= 'f') {
					*outp+=(*data)-'a'+0x0a;
				}
			} else data--; /* force loop exit */
		}
	} else {  /* no data - just add 10 bytes to size */
		outp+=10;
		if(outp-outpack > IP_MAXPACKET) 
			outp = outpack+IP_MAXPACKET;
	}

	/* then do any firther finalization (eg checksums) */
	if(icp) icmp_finalize(iph, icp, outp-outpack);
	if(udp) udp_finalize(iph, udp, outp-outpack);
	if(tcp) tcp_finalize(iph, tcp, outp-outpack);
	ip_finalize(iph, outp-outpack, to);

	/* Send the packet */
	sent = sendto(s, (char *)outpack, outp-outpack, 0, &whereto,
					  sizeof(struct sockaddr));
	if(sent == outp-outpack) {
		fprintf(stderr,"Sent %d bytes to %s\n",sent,hostname);
	} else {
      if (sent < 0)
			perror("sendto");
		else
			fprintf(stderr,"Only sent %d of %d bytes to %s\n",sent,outp-outpack,
					  hostname);
   }
	close(s);
	return 0;
}
